import json
import os
import subprocess
import urllib.parse
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional

from boltons.cacheutils import cachedproperty
from glom import glom
from glom import T
from glom.core import TType

from semgrep import __VERSION__
from semgrep.git import GIT_SH_TIMEOUT
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)


@dataclass
class GitMeta:
    """Gather metadata only from local filesystem."""

    cli_baseline_ref: Optional[str] = None
    environment: str = field(default="git", init=False)

    @property
    def event_name(self) -> str:
        if self.pr_id:
            return "pull_request"
        return "unknown"

    @property
    def repo_name(self) -> str:
        repo_name = os.getenv("SEMGREP_REPO_NAME")
        if repo_name:
            return repo_name

        rev_parse = subprocess.run(
            ["git", "rev-parse", "--show-toplevel"],
            capture_output=True,
            encoding="utf-8",
            timeout=GIT_SH_TIMEOUT,
        )
        if rev_parse.returncode != 0:
            raise Exception(
                "Unable to infer repo_name. Set SEMGREP_REPO_NAME environment variable or run in a valid git project"
            )

        repo_root_str = rev_parse.stdout.strip()
        return str(os.path.basename(repo_root_str))

    @property
    def repo_url(self) -> Optional[str]:
        return os.getenv("SEMGREP_REPO_URL")

    @property
    def commit_sha(self) -> Optional[str]:
        """
        Read commit hash of head from env var or run `git rev-parse HEAD`
        """
        commit = os.getenv("SEMGREP_COMMIT")
        if commit:
            return commit

        return subprocess.check_output(
            ["git", "rev-parse", "HEAD"], encoding="utf-8", timeout=GIT_SH_TIMEOUT
        ).strip()

    @property
    def head_ref(self) -> Optional[str]:
        return None

    @cachedproperty
    def base_commit_ref(self) -> Optional[str]:
        return self.cli_baseline_ref

    @property
    def ci_job_url(self) -> Optional[str]:
        return os.getenv("SEMGREP_JOB_URL")

    @property
    def pr_id(self) -> Optional[str]:
        return os.getenv("SEMGREP_PR_ID")

    @property
    def pr_title(self) -> Optional[str]:
        return os.getenv("SEMGREP_PR_TITLE")

    @property
    def branch(self) -> Optional[str]:
        branch = os.getenv("SEMGREP_BRANCH")
        if branch:
            return branch

        try:
            rev_parse = subprocess.run(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"],
                capture_output=True,
                encoding="utf-8",
                timeout=GIT_SH_TIMEOUT,
            )
            return rev_parse.stdout.strip()
        except Exception as e:
            logger.debug(f"Could not get branch name using git: {e}")
            return None

    def initialize_repo(self) -> None:
        return

    @property
    def commit_datetime(self) -> str:
        """
        Returns epoch time as str of head commit
        """
        return subprocess.check_output(
            ["git", "show", "-s", "--format=%ct"],
            stderr=subprocess.PIPE,
            encoding="utf-8",
            timeout=GIT_SH_TIMEOUT,
        ).strip()

    def to_dict(self) -> Dict[str, Any]:
        commit_title = subprocess.check_output(
            ["git", "show", "-s", "--format=%B"],
            stderr=subprocess.PIPE,
            encoding="utf-8",
            timeout=GIT_SH_TIMEOUT,
        ).strip()
        commit_author_email = subprocess.check_output(
            ["git", "show", "-s", "--format=%ae"],
            stderr=subprocess.PIPE,
            encoding="utf-8",
            timeout=GIT_SH_TIMEOUT,
        ).strip()
        commit_author_name = subprocess.check_output(
            ["git", "show", "-s", "--format=%an"],
            stderr=subprocess.PIPE,
            encoding="utf-8",
            timeout=GIT_SH_TIMEOUT,
        ).strip()

        return {
            "semgrep_version": __VERSION__,
            # REQUIRED for semgrep-app backend
            "repository": self.repo_name,
            #  OPTIONAL for semgrep-app backend
            "repo_url": self.repo_url,
            "branch": self.branch,
            "ci_job_url": self.ci_job_url,
            "commit": self.commit_sha,
            "commit_author_email": commit_author_email,
            "commit_author_name": commit_author_name,
            "commit_author_username": None,
            "commit_author_image_url": None,
            "commit_title": commit_title,
            "on": self.event_name,
            "pull_request_author_username": None,
            "pull_request_author_image_url": None,
            "pull_request_id": self.pr_id,
            "pull_request_title": self.pr_title,
            "scan_environment": self.environment,
            "is_full_scan": self.base_commit_ref == None,
        }


@dataclass
class GithubMeta(GitMeta):
    """Gather metadata from GitHub Actions."""

    environment: str = field(default="github-actions", init=False)
    # the last attempt will be 4**10 == 1048576 commits
    MAX_FETCH_ATTEMPT_COUNT: int = field(default=10, init=False)

    def glom_event(self, spec: TType) -> Any:
        return glom(self.event, spec, default=None)

    @property
    def event(self) -> Dict[str, Any]:
        value = os.getenv("GITHUB_EVENT_PATH")
        if value:
            return json.loads(Path(value).read_text())  # type: ignore
        return {}

    @property
    def is_pull_request_event(self) -> bool:
        """Return if running on a PR, even for variant types such as `pull_request_target`."""
        return self.event_name in {"pull_request", "pull_request_target"}

    @property
    def repo_name(self) -> str:
        repo_name = os.getenv("GITHUB_REPOSITORY")
        if repo_name:
            return repo_name
        else:
            raise Exception("Could not get repo_name when running in GithubAction")

    @property
    def repo_url(self) -> Optional[str]:
        server_url = os.getenv("GITHUB_SERVER_URL", "https://github.com")

        if self.repo_name:
            return f"{server_url}/{self.repo_name}"
        return None

    @property
    def commit_sha(self) -> Optional[str]:
        if self.is_pull_request_event:
            # https://github.community/t/github-sha-not-the-same-as-the-triggering-commit/18286/2
            return self.glom_event(T["pull_request"]["head"]["sha"])  # type: ignore
        if self.event_name == "push":
            return os.getenv("GITHUB_SHA")
        return super().commit_sha

    @property
    def head_ref(self) -> Optional[str]:
        if self.is_pull_request_event:
            return self.commit_sha
        else:
            return None

    @property
    def base_branch_tip(self) -> Optional[str]:
        return self.glom_event(T["pull_request"]["base"]["sha"])  # type: ignore

    def _find_branchoff_point(self, attempt_count: int = 0) -> str:
        """
        GithubActions is a shallow clone and the "base" that github sends
        is not the merge base. We must fetch and get the merge-base ourselves
        """
        # Should only be called if head_ref is defined
        assert self.head_ref is not None
        assert self.base_branch_tip is not None

        # fetch 0, 4, 16, 64, 256, 1024, ...
        fetch_depth = 4**attempt_count if attempt_count else 0
        if attempt_count > self.MAX_FETCH_ATTEMPT_COUNT:  # get all commits on last try
            fetch_depth = 2**31 - 1  # git expects a signed 32-bit integer

        logger.debug(
            f"Attempting to find merge base, attempt_count={attempt_count}, fetch_depth={fetch_depth}"
        )

        if attempt_count:
            process = subprocess.run(
                [
                    "git",
                    "fetch",
                    "origin",
                    "--depth",
                    str(fetch_depth),
                    self.base_branch_tip,
                ],
                check=True,
                capture_output=True,
                encoding="utf-8",
                timeout=GIT_SH_TIMEOUT,
            )
            logger.debug(
                f"Base branch fetch: args={process.args}, stdout={process.stdout}, stderr={process.stderr}"
            )
            process = subprocess.run(
                ["git", "fetch", "origin", "--depth", str(fetch_depth), self.head_ref],
                check=True,
                encoding="utf-8",
                capture_output=True,
                timeout=GIT_SH_TIMEOUT,
            )
            logger.debug(
                f"Head branch fetch: args={process.args}, stdout={process.stdout}, stderr={process.stderr}"
            )

        try:  # check if both branches connect to the yet-unknown branch-off point now
            process = subprocess.run(
                ["git", "merge-base", self.base_branch_tip, self.head_ref],
                encoding="utf-8",
                capture_output=True,
                check=True,
                timeout=GIT_SH_TIMEOUT,
            )
        except subprocess.CalledProcessError as e:
            output = e.stderr.strip()
            if (
                output  # output is empty when unable to find branch-off point
                and "Not a valid " not in output  # the error when a ref is missing
            ):
                raise Exception(f"Unexpected git merge-base error message: ({output})")

            if attempt_count >= self.MAX_FETCH_ATTEMPT_COUNT:
                raise Exception(
                    f"Could not find branch-off point between the baseline tip {self.base_branch_tip} and current head '{self.head_ref}' "
                )

            return self._find_branchoff_point(attempt_count + 1)
        else:
            logger.debug(
                f"Found merge base: args={process.args}, stdout={process.stdout}, stderr={process.stderr}"
            )
            return process.stdout.strip()

    @cachedproperty
    def base_commit_ref(self) -> Optional[str]:
        if self.cli_baseline_ref:
            return self.cli_baseline_ref
        if self.is_pull_request_event and self.head_ref is not None:
            return self._find_branchoff_point()
        return None

    @property
    def ci_job_url(self) -> Optional[str]:
        value = os.getenv("GITHUB_RUN_ID")
        if self.repo_url and value:
            return f"{self.repo_url}/actions/runs/{value}"
        return None

    @property
    def event_name(self) -> str:
        return os.getenv("GITHUB_EVENT_NAME", "unknown")

    @property
    def pr_id(self) -> Optional[str]:
        pr_id = self.glom_event(T["pull_request"]["number"])
        return str(pr_id) if pr_id else None

    @property
    def pr_title(self) -> Optional[str]:
        pr_title = self.glom_event(T["pull_request"]["title"])
        return str(pr_title) if pr_title else None

    def initialize_repo(self) -> None:
        if self.is_pull_request_event and self.head_ref is not None:
            self._find_branchoff_point()
        return

    def to_dict(self) -> Dict[str, Any]:
        return {
            **super().to_dict(),
            "branch": os.getenv("GITHUB_REF"),
            "commit_author_username": self.glom_event(T["sender"]["login"]),
            "commit_author_image_url": self.glom_event(T["sender"]["avatar_url"]),
            "pull_request_author_username": self.glom_event(
                T["pull_request"]["user"]["login"]
            ),
            "pull_request_author_image_url": self.glom_event(
                T["pull_request"]["user"]["avatar_url"]
            ),
        }


@dataclass
class GitlabMeta(GitMeta):
    """Gather metadata from GitLab 10.0+"""

    environment: str = field(default="gitlab-ci", init=False)

    @staticmethod
    def _fetch_branch_get_merge_base(branch_name: str, head_sha: str) -> str:
        """
        Return merge base of current head and head commit in branch_name

        Use Gitlab env vars to fetch target branch
        By default gitlab pipelines do a shallow clone

        Moved out to method so tests can mock this

        Because this is mocked it is not well tested. Use caution when modifying
        """
        parts = urllib.parse.urlsplit(os.environ["CI_MERGE_REQUEST_PROJECT_URL"])
        parts = parts._replace(
            netloc=f"gitlab-ci-token:{os.environ['CI_JOB_TOKEN']}@{parts.netloc}"
        )
        url = urllib.parse.urlunsplit(parts)
        subprocess.run(
            ["git", "fetch", url, branch_name],
            check=True,
            timeout=GIT_SH_TIMEOUT,
            capture_output=True,
        )

        base_sha = subprocess.check_output(
            ["git", "merge-base", "--all", head_sha, "FETCH_HEAD"],
            encoding="utf-8",
            timeout=GIT_SH_TIMEOUT,
        ).strip()
        return base_sha

    @property
    def repo_name(self) -> str:
        return os.getenv("CI_PROJECT_PATH", "[unknown]")

    @property
    def repo_url(self) -> Optional[str]:
        return os.getenv("CI_PROJECT_URL")

    @property
    def commit_sha(self) -> Optional[str]:
        return os.getenv("CI_COMMIT_SHA")

    @property
    def commit_ref(self) -> Optional[str]:
        return os.getenv("CI_COMMIT_REF_NAME")

    @cachedproperty
    def base_commit_ref(self) -> Optional[str]:
        if self.cli_baseline_ref:
            return self.cli_baseline_ref
        target_branch = os.getenv("CI_MERGE_REQUEST_TARGET_BRANCH_NAME")
        if not target_branch:
            return None

        head_sha = subprocess.check_output(
            ["git", "rev-parse", "HEAD"],
            encoding="utf-8",
            timeout=GIT_SH_TIMEOUT,
        ).strip()
        return self._fetch_branch_get_merge_base(target_branch, head_sha)

    @property
    def ci_job_url(self) -> Optional[str]:
        return os.getenv("CI_JOB_URL")

    @property
    def event_name(self) -> str:
        gitlab_event_name = os.getenv("CI_PIPELINE_SOURCE", "unknown")
        if gitlab_event_name in ["merge_request_event", "external_pull_request_event"]:
            return "pull_request"
        return gitlab_event_name

    @property
    def pr_id(self) -> Optional[str]:
        return os.getenv("CI_MERGE_REQUEST_IID")

    @property
    def start_sha(self) -> Optional[str]:
        return os.getenv("CI_MERGE_REQUEST_DIFF_BASE_SHA")

    @property
    def pr_title(self) -> Optional[str]:
        return os.getenv("CI_MERGE_REQUEST_TITLE")

    def to_dict(self) -> Dict[str, Any]:
        return {
            **super().to_dict(),
            "branch": self.commit_ref,
            "base_sha": self.base_commit_ref,
            "start_sha": self.start_sha,
        }


def generate_meta_from_environment(baseline_ref: Optional[str]) -> GitMeta:
    # https://help.github.com/en/actions/configuring-and-managing-workflows/using-environment-variables
    if os.getenv("GITHUB_ACTIONS") == "true":
        return GithubMeta(baseline_ref)

    # https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
    elif os.getenv("GITLAB_CI") == "true":
        return GitlabMeta(baseline_ref)

    else:
        return GitMeta(baseline_ref)



import json
import time
from io import StringIO
from os import environ
from pathlib import Path
from typing import Any
from typing import Collection
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Union

from semgrep import __VERSION__
from semgrep.app.metrics import metric_manager
from semgrep.autofix import apply_fixes
from semgrep.config_resolver import get_config
from semgrep.constants import DEFAULT_TIMEOUT
from semgrep.constants import OutputFormat
from semgrep.constants import RuleSeverity
from semgrep.core_runner import CoreRunner
from semgrep.error import FilesNotFoundError
from semgrep.error import MISSING_CONFIG_EXIT_CODE
from semgrep.error import SemgrepError
from semgrep.git import BaselineHandler
from semgrep.ignores import FileIgnore
from semgrep.ignores import IGNORE_FILE_NAME
from semgrep.ignores import Parser
from semgrep.nosemgrep import process_ignores
from semgrep.output import DEFAULT_SHOWN_SEVERITIES
from semgrep.output import OutputHandler
from semgrep.output import OutputSettings
from semgrep.profile_manager import ProfileManager
from semgrep.profiling import ProfilingData
from semgrep.project import get_project_url
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatchMap
from semgrep.semgrep_types import JOIN_MODE
from semgrep.target_manager import FileTargetingLog
from semgrep.target_manager import TargetManager
from semgrep.util import partition
from semgrep.util import unit_str
from semgrep.verbose_logging import getLogger


logger = getLogger(__name__)


def get_file_ignore() -> FileIgnore:
    TEMPLATES_DIR = Path(__file__).parent / "templates"
    workdir = Path.cwd()

    # Meant to be used only by semgrep-action
    if "SEMGREP_R2C_INTERNAL_EXPLICIT_SEMGREPIGNORE" in environ:
        semgrepignore_path = Path(
            environ["SEMGREP_R2C_INTERNAL_EXPLICIT_SEMGREPIGNORE"]
        ).resolve()
        logger.verbose("Using explicit semgrepignore file from environment variable")
    else:
        semgrepignore_path = Path(workdir / IGNORE_FILE_NAME)
        if not semgrepignore_path.is_file():
            logger.verbose(
                "No .semgrepignore found. Using default .semgrepignore rules. See the docs for the list of default ignores: https://semgrep.dev/docs/cli-usage/#ignoring-files"
            )
            semgrepignore_path = TEMPLATES_DIR / IGNORE_FILE_NAME
        else:
            logger.verbose("using path ignore rules from user provided .semgrepignore")

    with semgrepignore_path.open() as f:
        file_ignore = FileIgnore(
            base_path=workdir,
            patterns=Parser(workdir).parse(f),
        )

    return file_ignore


def invoke_semgrep(
    config: Path,
    targets: List[Path],
    output_settings: Optional[OutputSettings] = None,
    **kwargs: Any,
) -> Union[Dict[str, Any], str]:
    """
    Return Semgrep results of 'config' on 'targets' as a dict|str
    Uses default arguments of 'semgrep_main.main' unless overwritten with 'kwargs'
    """
    if output_settings is None:
        output_settings = OutputSettings(output_format=OutputFormat.JSON)

    StringIO()
    output_handler = OutputHandler(output_settings)
    (
        filtered_matches_by_rule,
        _,
        _,
        _,
        filtered_rules,
        profiler,
        profiling_data,
        shown_severities,
    ) = main(
        output_handler=output_handler,
        target=[str(t) for t in targets],
        pattern="",
        lang="",
        configs=[str(config)],
        **kwargs,
    )

    output_handler.rules = frozenset(filtered_rules)
    output_handler.rule_matches = [
        m for ms in filtered_matches_by_rule.values() for m in ms
    ]
    output_handler.profiler = profiler
    output_handler.profiling_data = profiling_data
    output_handler.severities = shown_severities

    return json.loads(output_handler._build_output())  # type: ignore


def run_rules(
    filtered_rules: List[Rule],
    target_manager: TargetManager,
    core_runner: CoreRunner,
    output_handler: OutputHandler,
    dump_command_for_core: bool,
    deep: bool,
) -> Tuple[RuleMatchMap, List[SemgrepError], Set[Path], ProfilingData,]:
    join_rules, rest_of_the_rules = partition(
        lambda rule: rule.mode == JOIN_MODE,
        filtered_rules,
    )
    dependency_aware_rules = [
        r for r in rest_of_the_rules if r.project_depends_on is not None
    ]
    dependency_only_rules, rest_of_the_rules = partition(
        lambda rule: not rule.should_run_on_semgrep_core, rest_of_the_rules
    )
    filtered_rules = rest_of_the_rules

    (
        rule_matches_by_rule,
        semgrep_errors,
        all_targets,
        profiling_data,
    ) = core_runner.invoke_semgrep(
        target_manager, filtered_rules, dump_command_for_core, deep
    )

    if join_rules:
        import semgrep.join_rule as join_rule

        for rule in join_rules:
            join_rule_matches, join_rule_errors = join_rule.run_join_rule(
                rule.raw, [target.path for target in target_manager.targets]
            )
            join_rule_matches_by_rule = {Rule.from_json(rule.raw): join_rule_matches}
            rule_matches_by_rule.update(join_rule_matches_by_rule)
            output_handler.handle_semgrep_errors(join_rule_errors)

    if len(dependency_aware_rules) > 0:
        from semgrep.dependency_aware_rule import run_dependency_aware_rule
        from dependencyparser.find_lockfiles import make_dependency_trie

        targets = [t.path for t in target_manager.targets]
        top_level_target_rooted = list(targets[0].parents)
        top_level_target: Path = (
            targets[0]
            if len(top_level_target_rooted) == 0
            else top_level_target_rooted[-1]
        )
        langs = [l for r in dependency_aware_rules for l in r.languages]
        dep_trie = make_dependency_trie(top_level_target, langs, target_manager)

        for rule in dependency_aware_rules:
            (dep_rule_matches, dep_rule_errors,) = run_dependency_aware_rule(
                rule_matches_by_rule.get(rule, []),
                rule,
                dep_trie,
            )
            rule_matches_by_rule[rule] = dep_rule_matches
            output_handler.handle_semgrep_errors(dep_rule_errors)

    return (
        rule_matches_by_rule,
        semgrep_errors,
        all_targets,
        profiling_data,
    )


def remove_matches_in_baseline(
    head_matches_by_rule: RuleMatchMap, baseline_matches_by_rule: RuleMatchMap
) -> RuleMatchMap:
    """
    Remove the matches in head_matches_by_rule that also occur in baseline_matches_by_rule
    """
    logger.verbose("Removing matches that exist in baseline scan")
    kept_matches_by_rule: RuleMatchMap = {}

    num_removed = 0

    for rule, matches in head_matches_by_rule.items():
        baseline_matches = {
            match.ci_unique_key for match in baseline_matches_by_rule.get(rule, [])
        }
        kept_matches_by_rule[rule] = [
            match for match in matches if match.ci_unique_key not in baseline_matches
        ]
        num_removed += len(matches) - len(kept_matches_by_rule[rule])

    logger.verbose(
        f"Removed {unit_str(num_removed, 'finding')} that were in baseline scan"
    )
    return kept_matches_by_rule


def main(
    *,
    dump_command_for_core: bool = False,
    deep: bool = False,
    output_handler: OutputHandler,
    target: Sequence[str],
    pattern: Optional[str],
    lang: Optional[str],
    configs: Sequence[str],
    no_rewrite_rule_ids: bool = False,
    jobs: int = 1,
    include: Optional[Sequence[str]] = None,
    exclude: Optional[Sequence[str]] = None,
    strict: bool = False,
    autofix: bool = False,
    replacement: Optional[str] = None,
    dryrun: bool = False,
    disable_nosem: bool = False,
    no_git_ignore: bool = False,
    timeout: int = DEFAULT_TIMEOUT,
    max_memory: int = 0,
    max_target_bytes: int = 0,
    timeout_threshold: int = 0,
    skip_unknown_extensions: bool = False,
    severity: Optional[Sequence[str]] = None,
    optimizations: str = "none",
    baseline_commit: Optional[str] = None,
) -> Tuple[
    RuleMatchMap,
    List[SemgrepError],
    Set[Path],
    FileTargetingLog,
    List[Rule],
    ProfileManager,
    ProfilingData,
    Collection[RuleSeverity],
]:
    logger.debug(f"semgrep version {__VERSION__}")
    if include is None:
        include = []

    if exclude is None:
        exclude = []

    project_url = get_project_url()
    profiler = ProfileManager()

    rule_start_time = time.time()
    configs_obj, config_errors = get_config(
        pattern, lang, configs, replacement=replacement, project_url=project_url
    )
    all_rules = configs_obj.get_rules(no_rewrite_rule_ids)
    profiler.save("config_time", rule_start_time)

    if not severity:
        shown_severities = DEFAULT_SHOWN_SEVERITIES
        filtered_rules = all_rules
    else:
        shown_severities = {RuleSeverity(s) for s in severity}
        filtered_rules = [rule for rule in all_rules if rule.severity.value in severity]

    output_handler.handle_semgrep_errors(config_errors)

    if config_errors and strict:
        raise SemgrepError(
            f"Ran with --strict and got {unit_str(len(config_errors), 'error')} while loading configs",
            code=MISSING_CONFIG_EXIT_CODE,
        )

    if not pattern:
        config_id_if_single = (
            list(configs_obj.valid.keys())[0] if len(configs_obj.valid) == 1 else ""
        )
        invalid_msg = (
            f"({unit_str(len(config_errors), 'invalid config file')})"
            if len(config_errors)
            else ""
        )
        logger.verbose(
            f"running {len(filtered_rules)} rules from {unit_str(len(configs_obj.valid), 'config')} {config_id_if_single} {invalid_msg}".strip()
        )
        if len(config_errors) > 0:
            raise SemgrepError(
                f"invalid configuration file found ({len(config_errors)} configs were invalid)",
                code=MISSING_CONFIG_EXIT_CODE,
            )
        if len(configs_obj.valid) == 0:
            raise SemgrepError(
                """No config given. Run with `--config auto` or see https://semgrep.dev/docs/running-rules/ for instructions on running with a specific config
""",
                code=MISSING_CONFIG_EXIT_CODE,
            )

    # Initialize baseline here to fail early on bad args
    baseline_handler = None
    if baseline_commit:
        try:
            baseline_handler = BaselineHandler(baseline_commit)
        # TODO better handling
        except Exception as e:
            raise SemgrepError(e)

    respect_git_ignore = not no_git_ignore
    try:
        target_manager = TargetManager(
            includes=include,
            excludes=exclude,
            max_target_bytes=max_target_bytes,
            target_strings=target,
            respect_git_ignore=respect_git_ignore,
            baseline_handler=baseline_handler,
            allow_unknown_extensions=not skip_unknown_extensions,
            file_ignore=get_file_ignore(),
        )
    except FilesNotFoundError as e:
        raise SemgrepError(e)

    core_start_time = time.time()
    core_runner = CoreRunner(
        jobs=jobs,
        timeout=timeout,
        max_memory=max_memory,
        timeout_threshold=timeout_threshold,
        optimizations=optimizations,
    )

    logger.verbose("Rules:")
    for ruleid in sorted(rule.id for rule in filtered_rules):
        logger.verbose(f"- {ruleid}")

    rule_matches_by_rule, semgrep_errors, all_targets, profiling_data = run_rules(
        filtered_rules,
        target_manager,
        core_runner,
        output_handler,
        dump_command_for_core,
        deep,
    )
    profiler.save("core_time", core_start_time)
    output_handler.handle_semgrep_errors(semgrep_errors)

    paths_with_matches = list(
        {match.path for matches in rule_matches_by_rule.values() for match in matches}
    )
    findings_count = sum(len(matches) for matches in rule_matches_by_rule.values())

    # Run baseline if needed
    if baseline_handler:
        logger.info(f"  Current version has {unit_str(findings_count, 'finding')}.")
        logger.info("")
        if not paths_with_matches:
            logger.info(
                "Skipping baseline scan, because there are no current findings."
            )
        elif not (set(paths_with_matches) - set(baseline_handler.status.added)):
            logger.info(
                "Skipping baseline scan, because all current findings are in files that didn't exist in the baseline commit."
            )
        else:
            logger.info(f"Switching repository to baseline commit '{baseline_commit}'.")
            baseline_handler.print_git_log()
            logger.info("")
            try:
                with baseline_handler.baseline_context():
                    baseline_target_manager = TargetManager(
                        # only include the paths that had a match
                        includes=[str(path) for path in paths_with_matches],
                        excludes=exclude,
                        max_target_bytes=max_target_bytes,
                        target_strings=target,
                        respect_git_ignore=respect_git_ignore,
                        allow_unknown_extensions=not skip_unknown_extensions,
                        file_ignore=get_file_ignore(),
                    )

                    (
                        baseline_rule_matches_by_rule,
                        baseline_semgrep_errors,
                        baseline_targets,
                        baseline_profiling_data,
                    ) = run_rules(
                        # only the rules that had a match
                        [
                            rule
                            for rule, matches in rule_matches_by_rule.items()
                            if matches
                        ],
                        baseline_target_manager,
                        core_runner,
                        output_handler,
                        dump_command_for_core,
                        deep,
                    )
                    rule_matches_by_rule = remove_matches_in_baseline(
                        rule_matches_by_rule, baseline_rule_matches_by_rule
                    )
                    output_handler.handle_semgrep_errors(baseline_semgrep_errors)
            except Exception as e:
                raise SemgrepError(e)

    ignores_start_time = time.time()
    keep_ignored = disable_nosem or output_handler.formatter.keep_ignores()
    filtered_matches_by_rule, nosem_errors, num_ignored_by_nosem = process_ignores(
        rule_matches_by_rule,
        keep_ignored,
        strict=strict,
    )
    profiler.save("ignores_time", ignores_start_time)
    output_handler.handle_semgrep_errors(nosem_errors)

    num_findings = sum(len(v) for v in filtered_matches_by_rule.values())
    profiler.save("total_time", rule_start_time)
    if metric_manager.is_enabled():
        error_types = list(e.semgrep_error_type() for e in semgrep_errors)

        metric_manager.set_project_hash(project_url)
        metric_manager.set_configs_hash(configs)
        metric_manager.set_rules_hash(filtered_rules)
        metric_manager.set_num_rules(len(filtered_rules))
        metric_manager.set_num_targets(len(all_targets))
        metric_manager.set_num_findings(num_findings)
        metric_manager.set_num_ignored(num_ignored_by_nosem)
        metric_manager.set_profiling_times(profiler.dump_stats())
        total_bytes_scanned = sum(t.stat().st_size for t in all_targets)
        metric_manager.set_total_bytes_scanned(total_bytes_scanned)
        metric_manager.set_errors(error_types)
        metric_manager.set_rules_with_findings(filtered_matches_by_rule)
        metric_manager.set_run_timings(
            profiling_data, list(all_targets), filtered_rules
        )

    if autofix:
        apply_fixes(filtered_matches_by_rule, dryrun)

    return (
        filtered_matches_by_rule,
        semgrep_errors,
        all_targets,
        target_manager.ignore_log,
        filtered_rules,
        profiler,
        profiling_data,
        shown_severities,
    )



from dataclasses import dataclass
from enum import Enum
from typing import Dict
from typing import List
from typing import Optional
from typing import Set

from semgrep.error import SemgrepError
from semgrep.semgrep_types import Language

KNOWN_HASH_ALGORITHMS: Set[str] = {
    "sha256",
    "sha512",
    "sha1",
    "gomod",
    # go.sum files use a non-standard hashing algorithm based on multiple uses of sha256 and conversion to base 64
}


class PackageManagers(str, Enum):
    NPM = "npm"
    PYPI = "pypi"
    GEM = "gem"
    GOMOD = "gomod"
    CARGO = "cargo"
    MAVEN = "maven"


@dataclass(eq=True, order=True, frozen=True)
class LockfileDependency:
    name: str
    version: str
    namespace: PackageManagers
    # map from hash algorithm to list of hashes that are ok
    allowed_hashes: Dict[str, List[str]]
    # sometimes the lockfile gives us a resolved URL, sweet
    resolved_url: Optional[List[str]] = None

    def __post_init__(self) -> None:
        for k in self.allowed_hashes.keys():
            assert (
                k in KNOWN_HASH_ALGORITHMS
            ), f"unknown hash type {k} not in {KNOWN_HASH_ALGORITHMS}"


def languages_to_namespaces(langs: List[Language]) -> Set[PackageManagers]:
    LANGUAGE_TO_NAMESPACE = {
        Language("python"): PackageManagers.PYPI,
        Language("js"): PackageManagers.NPM,
        Language("ts"): PackageManagers.NPM,
        Language("ruby"): PackageManagers.GEM,
        Language("go"): PackageManagers.GOMOD,
        Language("rust"): PackageManagers.CARGO,
        Language("java"): PackageManagers.MAVEN,
    }
    if Language("generic") in langs:
        return set(LANGUAGE_TO_NAMESPACE.values())

    namespaces = set()
    for lang in langs:
        if lang not in LANGUAGE_TO_NAMESPACE:
            raise SemgrepError(
                f"r2c-internal-project-depends-on does not support language {lang}"
            )
        namespaces.add(LANGUAGE_TO_NAMESPACE[lang])
    return namespaces


NAMESPACE_TO_LOCKFILES = {
    PackageManagers.PYPI: {"Pipfile.lock"},
    PackageManagers.NPM: {"package-lock.json", "yarn.lock"},
    PackageManagers.GEM: {"Gemfile.lock"},
    PackageManagers.GOMOD: {"go.sum"},
    PackageManagers.CARGO: {"cargo.lock"},
    PackageManagers.MAVEN: {"pom.xml"},
}



import sys
from enum import Enum
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import click

from semgrep.app import app_session
from semgrep.app import auth
from semgrep.commands.wrapper import handle_command_errors
from semgrep.config_resolver import get_config
from semgrep.constants import SEMGREP_URL
from semgrep.error import FATAL_EXIT_CODE
from semgrep.project import get_project_url
from semgrep.test import get_config_filenames
from semgrep.test import get_config_test_filenames
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)

SEMGREP_REGISTRY_UPLOAD_URL = f"{SEMGREP_URL}/api/registry/rule"
SEMGREP_REGISTRY_VIEW_URL = f"{SEMGREP_URL}/r/"
SEMGREP_SNIPPET_VIEW_URL = f"{SEMGREP_URL}/s/"


class VisibilityState(str, Enum):
    ORG_PRIVATE: str = "org_private"
    UNLISTED: str = "unlisted"
    PUBLIC: str = "public"


class VisibilityStateType(click.ParamType):
    name = "visibility_state"

    def get_metavar(self, param: click.Parameter) -> str:
        return "[org_private|unlisted|public]"

    def convert(
        self,
        value: Any,
        param: Optional["click.Parameter"],
        ctx: Optional["click.Context"],
    ) -> Any:
        if value is None:
            return None
        if isinstance(value, str):
            lower = value.lower()
            if lower == "org_private":
                return VisibilityState.ORG_PRIVATE
            if lower == "unlisted":
                return VisibilityState.UNLISTED
            if lower == "public":
                return VisibilityState.PUBLIC
        self.fail("expected 'org_private', 'unlisted', or 'public'")


VISIBILITY_STATE = VisibilityStateType()


def _get_test_code_for_config(
    target: Path,
) -> Tuple[List[Path], Dict[Path, List[Path]]]:
    config = target
    config_filenames = get_config_filenames(config)
    if len(config_filenames) == 0:
        config_filenames = [target]
    config_test_filenames = get_config_test_filenames(config, config_filenames, target)
    return config_filenames, config_test_filenames


@click.command()
@click.argument("target", nargs=1, type=click.Path(allow_dash=True))
@click.option(
    "--visibility",
    "visibility",
    default="org_private",
    type=VISIBILITY_STATE,
    help="Sets visibility of the uploaded rules."
    " If 'org_private', rules will be private to your org (default)"
    " If 'unlisted', rules will be listed in your org, but not listed to non-org members"
    " If 'public', rules will be published to the Semgrep Registry (requires --registry-id)",
)
@click.option(
    "--registry-id",
    "registry_id",
    help="If --visibility is set to public, this is the path the rule will have in the registry (example: python.flask.my-new-rule",
)
@handle_command_errors
def publish(
    target: str, visibility: VisibilityState, registry_id: Optional[str]
) -> None:
    """
    Upload rule to semgrep.dev

    Must be logged in to use; see `semgrep login`
    """
    if not app_session.token:
        click.echo("run `semgrep login` before using upload", err=True)
        sys.exit(FATAL_EXIT_CODE)

    fail_count = 0
    config_filenames, config_test_filenames = _get_test_code_for_config(Path(target))
    if len(config_filenames) == 0:
        click.echo(f"No valid Semgrep rules found in {target}", err=True)
        sys.exit(FATAL_EXIT_CODE)
    if len(config_filenames) != 1 and visibility == VisibilityState.PUBLIC:
        click.echo(
            f"Only one public rule can be uploaded at a time: specify a single Semgrep rule",
            err=True,
        )
        sys.exit(FATAL_EXIT_CODE)
    if visibility == VisibilityState.PUBLIC and registry_id is None:
        click.echo(f"--visibility=public requires --registry-id", err=True)
        sys.exit(FATAL_EXIT_CODE)

    click.echo(
        f'Found {len(config_filenames)} configs to publish with visibility "{visibility}"'
    )

    for config_filename in config_filenames:
        test_cases = config_test_filenames.get(config_filename, [])
        click.echo(
            f"--> Uploading {config_filename} (test cases: {[str(t) for t in test_cases]})"
        )
        first_test_case = test_cases[0] if len(test_cases) >= 1 else None

        if not _upload_rule(
            config_filename,
            visibility,
            first_test_case,
            registry_id,
        ):
            fail_count += 1
    if fail_count == 0:
        sys.exit(0)
    else:
        click.echo(f"{fail_count} rules failed to upload", err=True)
        sys.exit(FATAL_EXIT_CODE)


def _upload_rule(
    rule_file: Path,
    visibility: VisibilityState,
    test_code_file: Optional[Path],
    registry_id: Optional[str],
) -> bool:
    """
    Uploads rule in rule_file with the specificied visibility
    Args:
        rule_file: path to valid rule yaml file that contains single
        rule to be uploaded
        token: token with permissions to upload a rule
        visibility: the visibility of the uploaded rule
        test_code_file: optional test case to attach with the rule
    """
    config, errors = get_config(None, None, [str(rule_file)], project_url=None)

    if errors:
        click.echo(
            f"    Invalid rule definition: {str(rule_file)} is invalid: {errors}",
            err=True,
        )
        return False

    rules = config.get_rules(True)
    if len(rules) != 1:
        click.echo(
            "    Rule contains more than one rule: only yaml files with a single can be published",
            err=True,
        )
        return False

    rule = rules[0]

    # add metadata about the origin of the rule
    origin_note = f"published from {rule_file} in {get_project_url()}"
    rule.metadata["rule-origin-note"] = origin_note

    request_json = {
        "definition": {"rules": [rule._raw]},
        "visibility": visibility,
        # below should always be defined if passed validation
        "language": rule.languages[0] if len(rule.languages) >= 1 else None,
        # TODO backend can infer deployment ID from token; shouldn't need this
        "deployment_id": auth.get_deployment_id(),
        "test_target": test_code_file.read_text() if test_code_file else None,
        "registry_check_id": registry_id,
    }
    response = app_session.post(SEMGREP_REGISTRY_UPLOAD_URL, json=request_json)

    if not response.ok:
        click.echo(
            f"    Failed to upload rule with status_code {response.status_code}",
            err=True,
        )
        click.echo(response.text, err=True)
        return False
    else:
        created_rule = response.json()

        if visibility == VisibilityState.PUBLIC:
            click.echo(
                f"    Pull request created for this public rule at: {created_rule['pr_url']}"
            )
        elif visibility == VisibilityState.UNLISTED:
            click.echo(
                f"    Published {visibility} rule at {SEMGREP_SNIPPET_VIEW_URL}{created_rule['id']}"
            )
        else:
            click.echo(
                f"    Published {visibility} rule at {SEMGREP_REGISTRY_VIEW_URL}{created_rule['path']}"
            )

    return True



import os
import shutil
import stat
import subprocess
import sys
from pathlib import Path

import click
from tqdm import tqdm

from semgrep.app import app_session
from semgrep.commands.wrapper import handle_command_errors
from semgrep.constants import SEMGREP_URL
from semgrep.error import FATAL_EXIT_CODE
from semgrep.error import INVALID_API_KEY_EXIT_CODE
from semgrep.semgrep_core import SemgrepCore
from semgrep.util import set_flags
from semgrep.util import sub_check_output
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)


@click.command(hidden=True)
@handle_command_errors
def install_deep_semgrep() -> None:
    """
    Install the DeepSemgrep binary (Experimental)

    The binary is installed in the same directory that semgrep-core
    is installed in. Does not reinstall if existing binary already exists.

    Must be logged in and have access to DeepSemgrep beta
    Visit https://semgrep.dev/deep-semgrep-beta for more information
    """
    set_flags(verbose=False, debug=False, quiet=False, force_color=False)

    core_path = SemgrepCore.path()
    if core_path is None:
        logger.info(
            "Could not find `semgrep-core` executable so not sure where to install DeepSemgrep"
        )
        logger.info("There is something wrong with your semgrep installtation")
        sys.exit(FATAL_EXIT_CODE)

    deep_semgrep_path = Path(core_path).parent / "deep-semgrep"
    if deep_semgrep_path.exists():
        logger.info(
            f"Overwriting DeepSemgrep binary already installed in {deep_semgrep_path}"
        )

    if app_session.token is None:
        logger.info("run `semgrep login` before using `semgrep install`")
        sys.exit(INVALID_API_KEY_EXIT_CODE)

    if sys.platform.startswith("darwin"):
        platform = "osx"
    elif sys.platform.startswith("linux"):
        platform = "manylinux"
    else:
        platform = "manylinux"
        logger.info(
            "Running on potentially unsupported platform. Installing linux compatible binary"
        )

    url = f"{SEMGREP_URL}/api/agent/deployments/deepbinary/{platform}"

    with app_session.get(url, timeout=60, stream=True) as r:
        if r.status_code == 401:
            logger.info(
                "API token not valid. Try to run `semgrep logout` and `semgrep login` again."
            )
            sys.exit(INVALID_API_KEY_EXIT_CODE)
        if r.status_code == 403:
            logger.info("Logged in deployment does not have access to DeepSemgrep beta")
            logger.info(
                "Visit https://semgrep.dev/deep-semgrep-beta for more information."
            )
            sys.exit(FATAL_EXIT_CODE)
        r.raise_for_status()

        file_size = int(r.headers.get("Content-Length", 0))

        with open(deep_semgrep_path, "wb") as f:
            with tqdm.wrapattr(r.raw, "read", total=file_size) as r_raw:
                shutil.copyfileobj(r_raw, f)

    os.chmod(
        deep_semgrep_path,
        os.stat(deep_semgrep_path).st_mode | stat.S_IEXEC | stat.S_IXGRP | stat.S_IXOTH,
    )
    logger.info(f"Installed deepsemgrep to {deep_semgrep_path}")

    version = sub_check_output(
        [str(deep_semgrep_path), "--version"],
        timeout=10,
        encoding="utf-8",
        stderr=subprocess.DEVNULL,
    ).rstrip()
    logger.info(f"DeepSemgrep Version Info: ({version})")



# find lockfiles
from dataclasses import dataclass
from pathlib import Path
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional

from dependencyparser.models import languages_to_namespaces
from dependencyparser.models import LockfileDependency
from dependencyparser.models import NAMESPACE_TO_LOCKFILES
from dependencyparser.models import PackageManagers
from dependencyparser.parse_lockfile import LOCKFILE_PARSERS
from dependencyparser.parse_lockfile import parse_lockfile_str
from semgrep.semgrep_types import Language
from semgrep.target_manager import TargetManager

TARGET_LOCKFILE_FILENAMES = LOCKFILE_PARSERS.keys()


@dataclass
class Node:
    """
    Nodes of a DependencyTrie
    """

    children: Dict[str, "Node"]
    val: Optional[Dict[Path, List[LockfileDependency]]]


class DependencyTrie:
    """
    A trie of lockfile paths (root)
    Also contains a mapping from lockfile namespaces to mappings from lockfiles to their dependencies (all_deps)

    The path `A/B/lock1`, containing the parsed dependencies `deps1` is represented as

    A -> B -> {A/B/lock1 : deps1}

    If we insert `A/B/lock2` it becomes

    A -> B -> {A/B/lock1 : deps1, A/B/lock2 : deps2}

    If we insert `A/C/lock3` it becomes

    A -> B -> {A/B/lock1 : deps1, A/B/lock2 : deps2}
    |
     ->  C -> {A/C/lock3 : deps3}
    """

    def __init__(self) -> None:
        self.root: Node = Node(children={}, val=None)
        self.all_deps: Dict[PackageManagers, Dict[Path, List[LockfileDependency]]] = {}

    def insert(
        self, path: Path, deps: List[LockfileDependency], namespace: PackageManagers
    ) -> None:
        """
        Inserts `path` into the trie in accordance with the behavior described in the class comment
        `namespace` is namespace that this lockfile will be listed under in all_deps
        """
        if namespace in self.all_deps:
            self.all_deps[namespace][path] = deps
        else:
            self.all_deps[namespace] = {path: deps}

        curr = self.root
        for part in path.parts[:-1]:
            if part not in curr.children:
                curr.children[part] = Node(children={}, val=None)
            curr = curr.children[part]

        if curr.val is None:
            curr.val = {path: deps}
        else:
            curr.val[path] = deps

    def find_dependencies(
        self, path: Path
    ) -> Optional[Dict[Path, List[LockfileDependency]]]:
        """
        Finds all the lockfiles and their parsed dependencies that `path` depends on
        Traversing the trie finds the lockfile whose path shares the longest common
        prefix with `path`
        """
        curr = self.root
        for part in path.parts:
            if part not in curr.children:
                return curr.val
            else:
                curr = curr.children[part]
        return None


def make_dependency_trie(
    target: Path, langs: List[Language], target_manager: TargetManager
) -> DependencyTrie:
    dep_trie = DependencyTrie()
    # Triple for loop, but the outer two are (basically) constant time and guaranteed to be almost instant
    for namespace in languages_to_namespaces(langs):
        for lockfile_type in NAMESPACE_TO_LOCKFILES[namespace]:
            lockfiles: Iterable[Path] = target.glob("**/" + lockfile_type)
            lockfiles = (
                target_manager.file_ignore.filter_paths(candidates=lockfiles).kept
                if target_manager.file_ignore
                else lockfiles
            )
            for lockfile in lockfiles:
                deps = list(parse_lockfile_str(lockfile.read_text(), lockfile))
                dep_trie.insert(lockfile, deps, namespace)

    return dep_trie



import sys
from functools import wraps
from typing import Any
from typing import Callable

from semgrep import __VERSION__
from semgrep.app.metrics import metric_manager
from semgrep.error import FATAL_EXIT_CODE
from semgrep.error import OK_EXIT_CODE
from semgrep.error import SemgrepError
from semgrep.verbose_logging import getLogger


def handle_command_errors(func: Callable) -> Callable:
    """
    Adds the following functionality to our subcommands:
    - Enforces that exit code 1 is only for findings
    - Sets global logging level
    - Handles metric sending before exit

    This needed to be done in a decorator so testing
    using click.CliRunner would include this functionality.
    In particular putting this in __main__ (outside click),
    would have been harder to test
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> None:
        # When running semgrep as a command line tool
        # silence root level logger otherwise logs higher
        # than warning are handled twice
        logger = getLogger("semgrep")
        logger.propagate = False
        metric_manager.set_version(__VERSION__)

        try:
            func(*args, **kwargs)
        # Catch custom exception, output the right message and exit
        except SemgrepError as e:
            metric_manager.set_return_code(e.code)
            sys.exit(e.code)
        except Exception as e:
            logger.exception(e)
            metric_manager.set_return_code(FATAL_EXIT_CODE)
            sys.exit(FATAL_EXIT_CODE)
        else:
            metric_manager.set_return_code(OK_EXIT_CODE)
            sys.exit(OK_EXIT_CODE)
        finally:
            metric_manager.send()

    return wrapper



import binascii
import itertools
import textwrap
from collections import Counter
from datetime import datetime
from functools import total_ordering
from pathlib import Path
from typing import Any
from typing import Counter as CounterType
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING
from uuid import UUID

from attrs import evolve
from attrs import field
from attrs import frozen

import semgrep.output_from_core as core
from semgrep.constants import NOSEM_INLINE_COMMENT_RE
from semgrep.constants import RuleSeverity
from semgrep.external.pymmh3 import hash128  # type: ignore[attr-defined]

if TYPE_CHECKING:
    from semgrep.rule import Rule


def rstrip(value: Optional[str]) -> Optional[str]:
    return value.rstrip() if value else None


@total_ordering
@frozen(eq=False)
class RuleMatch:
    """
    A section of code that matches a single rule (which is potentially many patterns).

    This is also often referred to as a finding.
    TODO: Rename this class to Finding?
    """

    rule_id: str
    message: str = field(repr=False)
    severity: RuleSeverity

    path: Path = field(repr=str)
    start: core.Position
    end: core.Position

    metadata: Dict[str, Any] = field(repr=False, factory=dict)
    extra: Dict[str, Any] = field(repr=False, factory=dict)

    # We call rstrip() for consistency with semgrep-core, which ignores whitespace
    # including newline chars at the end of multiline patterns
    fix: Optional[str] = field(converter=rstrip, default=None)
    fix_regex: Optional[Dict[str, Any]] = None

    index: int = 0

    # None means we didn't check; ignore status is unknown
    is_ignored: Optional[bool] = field(default=None)

    # derived attributes
    lines: List[str] = field(init=False, repr=False)
    previous_line: str = field(init=False, repr=False)
    syntactic_context: str = field(init=False, repr=False)
    cli_unique_key: Tuple = field(init=False, repr=False)
    ci_unique_key: Tuple = field(init=False, repr=False)
    ordering_key: Tuple = field(init=False, repr=False)
    syntactic_id: str = field(init=False, repr=False)

    @lines.default
    def get_lines(self) -> List[str]:
        """
        Return lines in file that this RuleMatch is referring to.

        Assumes file exists.

        Need to do on initialization instead of on read since file might not be the same
        at read time
        """
        # Start and end line are one-indexed, but the subsequent slice call is
        # inclusive for start and exclusive for end, so only subtract from start
        start_line = self.start.line - 1
        end_line = self.end.line

        if start_line == -1 and end_line == 0:
            # Completely empty file
            return []

        # buffering=1 turns on line-level reads
        with self.path.open(buffering=1, errors="replace") as fd:
            result = list(itertools.islice(fd, start_line, end_line))

        return result

    @previous_line.default
    def get_previous_line(self) -> str:
        """Return the line preceding the match, if any.

        This is meant for checking for the presence of a nosemgrep comment.
        This implementation was derived from the 'lines' method below.
        Refer to it for relevant comments.
        IT feels like a lot of duplication. Feel free to improve.
        """
        # see comments in '_get_lines' method
        start_line = self.start.line - 2
        end_line = start_line + 1
        is_empty_file = self.end.line <= 0

        if start_line < 0 or is_empty_file:
            # no previous line
            return ""

        with self.path.open(buffering=1, errors="replace") as fd:
            res = list(itertools.islice(fd, start_line, end_line))

        if res:
            return res[0]
        else:
            return ""

    @syntactic_context.default
    def get_syntactic_context(self) -> str:
        """
        The code that matched, with whitespace and nosem comments removed.

        This is useful to so that findings can be considered the same
        when `    5 == 5` is updated to `  5 == 5  # nosemgrep`,
        and thus CI systems don't retrigger notifications.
        """
        lines = [*self.lines]
        if len(lines) > 0:
            lines[0] = NOSEM_INLINE_COMMENT_RE.sub("", lines[0])
            lines[0] = lines[0].rstrip() + "\n"

        code = "".join(lines)  # the lines end with newlines already
        code = textwrap.dedent(code)
        code = code.strip()
        return code

    @cli_unique_key.default
    def get_cli_unique_key(self) -> Tuple:
        """
        A unique key designed with data-completeness & correctness in mind.

        Results in more unique findings than ci_unique_key.

        Used for deduplication in the CLI before writing output.
        """
        return (
            self.rule_id,
            str(self.path),
            self.start.offset,
            self.end.offset,
            self.message,
        )

    @ci_unique_key.default
    def get_ci_unique_key(self) -> Tuple:
        """
        A unique key designed with notification user experience in mind.

        Results in fewer unique findings than cli_unique_key.

        We use this to check if a finding matches its baseline equivalent,
        and to de-duplicate findings when pushing to semgrep.dev from CI,
        so that you don't get multiple notifications for the same finding
        when just moving it around commit-after-commit in a pull request.

        Some things that this key deduplicates to reduce useless notifications:
        - findings that match different metavariables within the same code snippet
        - findings that differ only in indentation
        - findings that differ only because one is ignored with `# nosemgrep`

        This key was originally implemented in and ported from semgrep-agent.
        """
        try:
            path = self.path.relative_to(Path.cwd())
        except ValueError:
            path = self.path
        return (self.rule_id, str(path), self.syntactic_context, self.index)

    @ordering_key.default
    def get_ordering_key(self) -> Tuple:
        """
        Used to sort findings in output.

        Note that we often batch by rule ID when gathering matches,
        so the included self.rule_id will not do anything in those cases.

        The message field is included to ensure a consistent ordering
        when two findings match with different metavariables on the same code.
        """
        return (self.path, self.start, self.end, self.rule_id, self.message)

    @syntactic_id.default
    def get_syntactic_id(self) -> str:
        """
        A 32-character hash representation of ci_unique_key.

        This value is sent to semgrep.dev and used as a unique key in the database.
        """
        # Upon reviewing an old decision,
        # there's no good reason for us to use MurmurHash3 here,
        # but we need to keep consistent hashes so we cannot change this easily
        hash_int = hash128(str(self.ci_unique_key))
        hash_bytes = int.to_bytes(hash_int, byteorder="big", length=16, signed=False)
        return str(binascii.hexlify(hash_bytes), "ascii")

    @property
    def uuid(self) -> UUID:
        """
        A UUID representation of ci_unique_key.
        """
        return UUID(hex=self.syntactic_id)

    @property
    def is_blocking(self) -> bool:
        """
        Returns if this finding indicates it should block CI
        """
        return "block" in self.metadata.get("dev.semgrep.actions", ["block"])

    def to_app_finding_format(self, commit_date: str) -> Dict[str, Any]:
        """
        commit_date here for legacy reasons.
        commit date of the head commit in epoch time
        """
        commit_date_app_format = datetime.fromtimestamp(int(commit_date)).isoformat()

        # Follow semgrep.dev severity conventions
        if self.severity.value == RuleSeverity.ERROR.value:
            app_severity = 2
        elif self.severity.value == RuleSeverity.WARNING.value:
            app_severity = 1
        else:
            app_severity = 0

        ret = {
            "check_id": self.rule_id,
            "path": str(self.path),
            "line": self.start.line,
            "column": self.start.col,
            "end_line": self.end.line,
            "end_column": self.end.col,
            "message": self.message,
            "severity": app_severity,
            "index": self.index,
            "commit_date": commit_date_app_format,
            "syntactic_id": self.syntactic_id,
            "metadata": self.metadata,
            "is_blocking": self.is_blocking,
        }

        if self.extra.get("fixed_lines"):
            ret["fixed_lines"] = self.extra.get("fixed_lines")
        return ret

    def __hash__(self) -> int:
        """
        We use the "data-correctness" key to prevent keeping around duplicates.
        """
        return hash(self.cli_unique_key)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.cli_unique_key == other.cli_unique_key

    def __lt__(self, other: "RuleMatch") -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.ordering_key < other.ordering_key


class RuleMatchSet(Set[RuleMatch]):
    """
    A custom set type which is aware when findings are the same.

    It also automagically adds the correct finding index when adding elements.
    """

    def __init__(self, __iterable: Optional[Iterable[RuleMatch]] = None) -> None:
        self._ci_key_counts: CounterType[Tuple] = Counter()
        if __iterable is None:
            super().__init__()
        else:
            super().__init__(__iterable)

    def add(self, match: RuleMatch) -> None:
        """
        Add finding, but if the same (rule, path, code) exists,
        note this by incrementing the finding's index.

        The index lets us still notify when some code with findings is duplicated,
        even though we'd otherwise deduplicate the findings.
        """
        self._ci_key_counts[match.ci_unique_key] += 1
        match = evolve(match, index=self._ci_key_counts[match.ci_unique_key] - 1)
        super().add(match)

    def update(self, *rule_match_iterables: Iterable[RuleMatch]) -> None:
        """
        Add findings, but if the same (rule, path, code) exists,
        note this by incrementing the finding's index.

        The index lets us still notify when some code with findings is duplicated,
        even though we'd otherwise deduplicate the findings.
        """
        for rule_matches in rule_match_iterables:
            for rule_match in rule_matches:
                self.add(rule_match)


# Our code orders findings at one point and then just assumes they're in order.
# This type marks variables that went through ordering already.
OrderedRuleMatchList = List[RuleMatch]
RuleMatchMap = Dict["Rule", OrderedRuleMatchList]
